import flax.linen as nn


class VGGBlock(nn.Module):
    num_features: int = 64
    num_convs: int = 2
    kernel_initializer: nn.initializers = nn.initializers.lecun_normal()
    bias_initializer: nn.initializers = nn.initializers.zeros

    @nn.compact
    def __call__(self, x):
        for _ in range(self.num_convs):
            x = nn.Conv(
                self.num_features,
                (3, 3),
                padding="same",
                kernel_init=self.kernel_initializer,
                bias_init=self.bias_initializer,
            )(x)
            x = nn.relu(x)
        x = nn.max_pool(x, window_shape=(2, 2), strides=(2, 2))
        return x



import jax.numpy as jnp
import flax.linen as nn

from .blocks import VGGBlock


class VGG19(nn.Module):
    include_top: bool = True
    pooling: str = "avg"
    num_classes: int = 1000
    kernel_initializer: nn.initializers = nn.initializers.lecun_normal()
    bias_initializer: nn.initializers = nn.initializers.zeros
    classifier_activation: nn.Module = nn.softmax

    @nn.compact
    def __call__(self, x):
        # Block 1
        x = VGGBlock(
            num_features=64,
            kernel_initializer=self.bias_initializer,
            bias_initializer=self.bias_initializer,
        )(x)
        # Block 2
        x = VGGBlock(
            num_features=128,
            kernel_initializer=self.bias_initializer,
            bias_initializer=self.bias_initializer,
        )(x)
        # Block 3
        x = VGGBlock(
            num_features=256,
            num_convs=4,
            kernel_initializer=self.bias_initializer,
            bias_initializer=self.bias_initializer,
        )(x)
        # Block 4
        x = VGGBlock(
            num_features=512,
            num_convs=4,
            kernel_initializer=self.bias_initializer,
            bias_initializer=self.bias_initializer,
        )(x)
        # Block 5
        x = VGGBlock(
            num_features=512,
            num_convs=4,
            kernel_initializer=self.bias_initializer,
            bias_initializer=self.bias_initializer,
        )(x)
        if self.include_top:
            x = jnp.reshape(x, (x.shape[0], -1))  # Flatten
            x = nn.Dense(
                4096,
                kernel_init=self.bias_initializer,
                bias_init=self.bias_initializer,
            )(x)
            x = nn.relu(x)
            x = nn.Dense(
                4096,
                kernel_init=self.bias_initializer,
                bias_init=self.bias_initializer,
            )(x)
            x = nn.relu(x)
            x = nn.Dense(
                self.num_classes,
                kernel_init=self.bias_initializer,
                bias_init=self.bias_initializer,
            )(x)
            x = self.classifier_activation(x)
        else:
            if self.pooling == "avg":
                x = jnp.mean(x, axis=(1, 2))  # Global Average Pooling
            elif self.pooling == "max":
                x = jnp.max(x, axis=(1, 2))  # Global Max Pooling
        return x



import os
import re
import setuptools


def find_version(file_path):
    with open(file_path, "r", encoding="utf-8") as fh:
        file_content = fh.read()
    version_match = re.search(r'^__version__ = ["\']([^"\']*)["\']', file_content, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    dependencies = f.read().splitlines()


setuptools.setup(
    name="flax-vision-models",
    version=find_version(os.path.join('flax_vision_models', '__init__.py')),
    author="Soumik Rakshit",
    author_email="19soumik.rakshit96@gmail.com",
    description="A repository of Deep Learning models in Flax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/soumik12345/flax-vision-models",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(exclude=("notebooks/")),
    python_requires=">=3.7.12,<3.11",
    install_requires=dependencies,
)


import jax.numpy as jnp

from typing import List, Optional


def preprocess(
    x,
    mean: Optional[List[float]] = None,
    std: Optional[float] = None,
    dtype=jnp.float64,
):
    """ "Reference: https://github.com/keras-team/keras/blob/d8fcb9d4d4dad45080ecfdd575483653028f8eda/keras/applications/imagenet_utils.py#L168"""
    x = x.astype(dtype)
    x = x[..., ::-1]
    mean = [103.939, 116.779, 123.68] if mean is None else mean
    x[..., 0] -= mean[0]
    x[..., 1] -= mean[1]
    x[..., 2] -= mean[2]
    if std is not None:
        x[..., 0] /= std[0]
        x[..., 1] /= std[1]
        x[..., 2] /= std[2]
    return x



import os
import wget
import json
import wandb


CATEGORIES_ARTIFACT_ADDRESS = (
    "geekyrakshit/flax-vision-models/imagenet-simple-labels:v0"
)
CATEGORIES_FILE_ADDRESS = "imagenet-simple-labels.json"

CATEGORIES_FILE_URL = "https://raw.githubusercontent.com/anishathalye/imagenet-simple-labels/master/imagenet-simple-labels.json"


def decode_probabilities_imagenet(topk_classes, topk_probabilities):
    imagenet_labels_file = None
    if wandb.run is not None:
        artifact = wandb.use_artifact(CATEGORIES_ARTIFACT_ADDRESS)
        artifact_dir = artifact.download()
        imagenet_labels_file = os.path.join(artifact_dir, CATEGORIES_FILE_ADDRESS)
    else:
        imagenet_labels_file = wget.download(CATEGORIES_FILE_URL)
    categories = json.load(open(imagenet_labels_file))
    topk_labels = [categories[topk_classes[i]] for i in range(topk_classes.shape[0])]
    topk_probabilities = [topk_probabilities[i] for i in range(topk_classes.shape[0])]
    if wandb.run is not None:
        wandb.log(
            {
                "Prediction": wandb.plot.bar(
                    wandb.Table(
                        data=[
                            [label, val]
                            for (label, val) in zip(topk_labels, topk_probabilities)
                        ],
                        columns=["Predicted Label", "Probability"],
                    ),
                    "Predicted Label",
                    "Probability",
                    title="Prediction",
                )
            }
        )
    return topk_labels, topk_probabilities



import os
import wget
import wandb
import deepdish
from typing import Tuple

import jax
import numpy as np
from clu import parameter_overview

from .vgg16 import VGG16
from .vgg19 import VGG19


VGG16_ARTIFACT_ADDRESS = "geekyrakshit/flax-vision-models/vgg16-keras-imagenet:v0"
VGG16_URL = "https://github.com/soumik12345/flax-vision-models/releases/download/0.0.1/vgg16_keras_imagenet.h5"
VGG16_FILE_NAME = "vgg16_keras_imagenet.h5"

VGG19_ARTIFACT_ADDRESS = "geekyrakshit/flax-vision-models/vgg19-keras-imagenet:v0"
VGG19_URL = "https://github.com/soumik12345/flax-vision-models/releases/download/0.0.1/vgg19_keras_imagenet.h5"
VGG19_FILE_NAME = "vgg19_keras_imagenet.h5"


def build_vgg16(
    input_shape: Tuple = (224, 224, 3),
    seed: int = 0,
    pretrained: bool = True,
    show_parameter_overview: bool = False,
    **kwargs
):
    key = jax.random.PRNGKey(seed)
    model = VGG16(**kwargs)
    parameters = model.init(key, np.random.randn(1, *input_shape))

    if pretrained:
        weight_file = None
        
        if wandb.run is not None:
            artifact = wandb.use_artifact(VGG16_ARTIFACT_ADDRESS)
            artifact_dir = artifact.download()
            weight_file = os.path.join(artifact_dir, VGG16_FILE_NAME)
        else:
            weight_file = wget.download(VGG16_URL)
        
        parameters = deepdish.io.load(weight_file)

    if show_parameter_overview:
        print(parameter_overview.get_parameter_overview(parameters))
    return model, parameters


def build_vgg19(
    input_shape: Tuple = (224, 224, 3),
    seed: int = 0,
    pretrained: bool = True,
    show_parameter_overview: bool = False,
    **kwargs
):
    key = jax.random.PRNGKey(seed)
    model = VGG19(**kwargs)
    parameters = model.init(key, np.random.randn(1, *input_shape))

    if pretrained:
        weight_file = None
        
        if wandb.run is not None:
            artifact = wandb.use_artifact(VGG19_ARTIFACT_ADDRESS)
            artifact_dir = artifact.download()
            weight_file = os.path.join(artifact_dir, VGG19_FILE_NAME)
        else:
            weight_file = wget.download(VGG19_URL)
        
        parameters = deepdish.io.load(weight_file)

    if show_parameter_overview:
        print(parameter_overview.get_parameter_overview(parameters))
    return model, parameters
